#include "STM32F767.h"

volatile RCC_T *RCC = (RCC_T*)0x40023800;
volatile GPIO_T *GPIOA = (GPIO_T*)0x40020000;
volatile GPIO_T *GPIOB = (GPIO_T*)0x40020400;
volatile GPIO_T *GPIOC = (GPIO_T*)0x40020800;
volatile GPIO_T *GPIOD = (GPIO_T*)0x40020C00;
volatile GPIO_T *GPIOE = (GPIO_T*)0x40021000;
volatile GPIO_T *GPIOF = (GPIO_T*)0x40021400;
volatile GPIO_T *GPIOG = (GPIO_T*)0x40021800;
volatile NVIC_T *NVIC = (NVIC_T*)0xE000E100;
volatile uint32_t *NVIC_STIR = (uint32_t*)0xE000EF00;
volatile FLASH_T *FLASH = (FLASH_T*)0x40023C00;
volatile DAC_T *DAC = (DAC_T*)0x40007400;
volatile TIM_2_5_T *TIM2 = (TIM_2_5_T*)0x40000000;
volatile ADC_T *ADC1 = (ADC_T*)0x40012000;
volatile ADC_T *ADC2 = (ADC_T*)0x40012100;
volatile ADC_T *ADC3 = (ADC_T*)0x40012200;
volatile ADC_COMMON_T *ADC_COMMON = (ADC_COMMON_T*)0x40012300;
volatile DMA_T *DMA1 = (DMA_T*)0x40026000;
volatile DMA_T *DMA2 = (DMA_T*)0x40026400;
volatile USART_T *USART1 = (USART_T*)0x40011000;
volatile USART_T *USART2 = (USART_T*)0x40004400;
volatile USART_T *USART3 = (USART_T*)0x40004800;
volatile ETH_T *ETH = (ETH_T*)0x40028000;
volatile SYSCFG_T *SYSCFG = (SYSCFG_T*)0x40013800;

void wait_cycles(int N)
{
    volatile static int c = 0;
    for(int i = 0; i < N; i++)
    {
        c++;
    }
}

void setup_clock_HSI(uint32_t N, uint32_t M, uint32_t P, uint32_t Q, uint32_t R)
{
    RCC->CR |= (0x1 << 0);    // HSI on
    
    // Wait for HSI to be ready
    while(((RCC->CR >> 1) & 0x1) == 0);
    
    // pre-dividers
    RCC->CFGR &= ~ (0x7 << 10);
    RCC->CFGR |= (6 << 10);     // PRE1 = 6 ==> div by 8 ==> 27MHz
    RCC->CFGR &= ~ (0x7 << 13);
    RCC->CFGR |= (6 << 13);     // PRE2 = 6 ==> div by 8 ==> 27MHz
    RCC->CFGR &= ~ (0xF << 4);
    RCC->CFGR &= ~(0xF << 4);   // HPRE = 0 ==> div by 1 ==> 216MHz = HCLK
    FLASH->ACR &= ~ (0xF << 0);
    FLASH->ACR |= (10 << 0);    // 10 wait states
    
    // Configure PLL
    // reset value of RCC->PLLCFGR = 0x24003010 = 0b 0010 0100 0000 0000 0011 0000 0001 0000
    RCC->PLLCFGR &= ~(0x1FF << 6);
    RCC->PLLCFGR |= ((N & 0x1FF) << 6);
    RCC->PLLCFGR &= ~(0x3F << 0);
    RCC->PLLCFGR |= ((M & 0x3F) << 0);
    RCC->PLLCFGR &= ~(0x3 << 16);
    RCC->PLLCFGR |= ((((P/2)-1) & 0x3) << 16);
    RCC->PLLCFGR &= ~(0xF << 24);
    RCC->PLLCFGR |= ((Q & 0xF) << 24);
    RCC->PLLCFGR &= ~(0x7 << 28);
    RCC->PLLCFGR |= ((R & 0x7) << 28);
    RCC->PLLCFGR &= ~(0x1 << 22);    // Set PLL clock source = HSI
    
    if(((RCC->PLLCFGR >> 6) & 0x1FF) != N) while(1);
    if(((RCC->PLLCFGR >> 0) & 0x3F) != M) while(1);
    if(((RCC->PLLCFGR >> 16) & 0x3) != 0) while(1);
    
    RCC->CR |= (0x1 << 24);    // PLL on
    
    // Wait for PLL to be ready
    while(((RCC->CR >> 25) & 0x1) == 0);
    
    RCC->CFGR &= ~(0x3 << 0);
    RCC->CFGR |= (0x2 << 0);    // Use PLL as system clock source
}

void setup_clock_HSE_bypass(uint32_t N, uint32_t M, uint32_t P, uint32_t Q, uint32_t R)
{
    RCC->CR |= (0x1 << 16);    // HSE on
    RCC->CR |= (0x1 << 18);    // HSE bypass (external clock)
    
    // Wait for HSE to be ready
    while(((RCC->CR >> 17) & 0x1) == 0);
    
    // pre-dividers
    RCC->CFGR &= ~ (0x7 << 10);
    RCC->CFGR |= (6 << 10);     // PRE1 = 6 ==> div by 8 ==> 27MHz
    RCC->CFGR &= ~ (0x7 << 13);
    RCC->CFGR |= (6 << 13);     // PRE2 = 6 ==> div by 8 ==> 27MHz
    RCC->CFGR &= ~ (0xF << 4);
    RCC->CFGR &= ~(0xF << 4);   // HPRE = 0 ==> div by 1 ==> 216MHz = HCLK
    FLASH->ACR &= ~ (0xF << 0);
    FLASH->ACR |= (10 << 0);    // 10 wait states
    
    // Configure PLL
    // reset value of RCC->PLLCFGR = 0x24003010 = 0b 0010 0100 0000 0000 0011 0000 0001 0000
    RCC->PLLCFGR &= ~(0x1FF << 6);
    RCC->PLLCFGR |= ((N & 0x1FF) << 6);
    RCC->PLLCFGR &= ~(0x3F << 0);
    RCC->PLLCFGR |= ((M & 0x3F) << 0);
    RCC->PLLCFGR &= ~(0x3 << 16);
    RCC->PLLCFGR |= ((((P/2)-1) & 0x3) << 16);
    RCC->PLLCFGR &= ~(0xF << 24);
    RCC->PLLCFGR |= ((Q & 0xF) << 24);
    RCC->PLLCFGR &= ~(0x7 << 28);
    RCC->PLLCFGR |= ((R & 0x7) << 28);
    RCC->PLLCFGR |= (0x01 << 22);    // Set PLL clock source = HSE
    
    if(((RCC->PLLCFGR >> 6) & 0x1FF) != N) while(1);
    if(((RCC->PLLCFGR >> 0) & 0x3F) != M) while(1);
    if(((RCC->PLLCFGR >> 16) & 0x3) != 0) while(1);
    
    RCC->CR |= (0x1 << 24);    // PLL on
    
    // Wait for PLL to be ready
    while(((RCC->CR >> 25) & 0x1) == 0);
    
    RCC->CFGR &= ~(0x3 << 0);
    RCC->CFGR |= (0x2 << 0);    // Use PLL as system clock source
}

void initGPIO()
{
    // Enable Port A, B, C, D, E, F and G
    RCC->AHB1ENR |= (0x01 << 6) | (0x01 << 5) | (0x01 << 4) | (0x01 << 3) | (0x01 << 2) | (0x01 << 1) | (0x01 << 0);
}

uint8_t getPinState(uint8_t port, uint8_t pin)
{
    uint8_t ret = -1;
    switch(port)
    {
        case PORTA:
            ret = ((GPIOA->IDR) >> pin) & 0x01;
        break;
        case PORTB:
            ret = ((GPIOB->IDR) >> pin) & 0x01;
        break;
        case PORTC:
            ret = ((GPIOC->IDR) >> pin) & 0x01;
        break;
        case PORTD:
            ret = ((GPIOD->IDR) >> pin) & 0x01;
        break;
        case PORTE:
            ret = ((GPIOE->IDR) >> pin) & 0x01;
        break;
        case PORTF:
            ret = ((GPIOF->IDR) >> pin) & 0x01;
        break;
        case PORTG:
            ret = ((GPIOG->IDR) >> pin) & 0x01;
        break;
    }
    
    return ret;
}

void setPinState(uint8_t port, uint8_t pin, uint8_t state)
{
    switch(port)
    {
        case PORTA:
            if(state == 1)
            {
                GPIOA->BSRR = (0x01 << pin);
            }
            else
            {
                GPIOA->BSRR = (0x01 << (pin + 16));
            }
        break;
        case PORTB:
            if(state == 1)
            {
                GPIOB->BSRR = (0x01 << pin);
            }
            else
            {
                GPIOB->BSRR = (0x01 << (pin + 16));
            }
        break;
        case PORTC:
            if(state == 1)
            {
                GPIOC->BSRR = (0x01 << pin);
            }
            else
            {
                GPIOC->BSRR = (0x01 << (pin + 16));
            }
        break;
        case PORTD:
            if(state == 1)
            {
                GPIOD->BSRR = (0x01 << pin);
            }
            else
            {
                GPIOD->BSRR = (0x01 << (pin + 16));
            }
        break;
        case PORTE:
            if(state == 1)
            {
                GPIOE->BSRR = (0x01 << pin);
            }
            else
            {
                GPIOE->BSRR = (0x01 << (pin + 16));
            }
        break;
        case PORTF:
            if(state == 1)
            {
                GPIOF->BSRR = (0x01 << pin);
            }
            else
            {
                GPIOF->BSRR = (0x01 << (pin + 16));
            }
        break;
        case PORTG:
            if(state == 1)
            {
                GPIOG->BSRR = (0x01 << pin);
            }
            else
            {
                GPIOG->BSRR = (0x01 << (pin + 16));
            }
        break;
    }
}

void setPinDir(uint8_t port, uint8_t pin, uint8_t dir)
{
    setPinDirExt(port, pin, dir, GPIO_PULLUP);
}

void setPinDirExt(uint8_t port, uint8_t pin, uint8_t dir, uint8_t pullup)
{
    volatile GPIO_T *gpio_reg;
    switch(port)
    {
        case PORTA:
            gpio_reg = GPIOA;
        break;
        case PORTB:
            gpio_reg = GPIOB;
        break;
        case PORTC:
            gpio_reg = GPIOC;
        break;
        case PORTD:
            gpio_reg = GPIOD;
        break;
        case PORTE:
            gpio_reg = GPIOE;
        break;
        case PORTF:
            gpio_reg = GPIOF;
        break;
        case PORTG:
            gpio_reg = GPIOG;
        break;
        default:
            gpio_reg = 0;
        break;
    }
    
    if(dir == GPIO_OUTPUT)
    {
        gpio_reg->MODER &= ~(0x03 << (2 * pin));
        gpio_reg->MODER |= (0x01 << (2 * pin));
        gpio_reg->OTYPER &= ~(0x01 << pin);
        gpio_reg->OSPEEDR &= ~(0x03 << (2 * pin));
        gpio_reg->OSPEEDR |= (0x01 << (2 * pin));
        gpio_reg->PUPDR &= ~(0x03 << (2 * pin));
    }
    else
    {
        gpio_reg->MODER &= ~(0x03 << (2 * pin));
        gpio_reg->OTYPER &= ~(0x01 << pin);
        gpio_reg->OSPEEDR &= ~(0x03 << (2 * pin));
        gpio_reg->OSPEEDR |= (0x01 << (2 * pin));
        gpio_reg->PUPDR &= ~(0x03 << (2 * pin));
        
        if(pullup == GPIO_PULLUP)
        {
            gpio_reg->PUPDR |= (0x01 << (2 * pin));
        }
    }
}

void setAlternateFunction(uint8_t port, uint8_t pin, uint8_t AF)
{
    volatile GPIO_T *gpio_reg;
    switch(port)
    {
        case PORTA:
            gpio_reg = GPIOA;
        break;
        case PORTB:
            gpio_reg = GPIOB;
        break;
        case PORTC:
            gpio_reg = GPIOC;
        break;
        case PORTD:
            gpio_reg = GPIOD;
        break;
        case PORTE:
            gpio_reg = GPIOE;
        break;
        case PORTF:
            gpio_reg = GPIOF;
        break;
        case PORTG:
            gpio_reg = GPIOG;
        break;
        default:
            gpio_reg = 0;
        break;
    }
    
    gpio_reg->MODER &= ~(0x03 << (2 * pin));
    gpio_reg->MODER |= (0x02 << (2 * pin));
    gpio_reg->OTYPER &= ~(0x01 << pin);
    gpio_reg->OSPEEDR &= ~(0x03 << (2 * pin));
    gpio_reg->OSPEEDR |= (0x03 << (2 * pin));
    //gpio_reg->PUPDR |= (0x01 << (2 * pin));
    gpio_reg->PUPDR &= ~(0x03 << (2 * pin));
    
    if(pin < 8)
    {
        gpio_reg->AFRL &= ~(0xF << (4 * pin));
        gpio_reg->AFRL |= (AF << (4 * pin));
    }
    else
    {
        gpio_reg->AFRH &= ~(0xF << (4 * (pin - 8)));
        gpio_reg->AFRH |= (AF << (4 * (pin - 8)));
    }
}
