#include "UART.h"

volatile uint8_t hex_chars[] = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};

void init_UART()
{
    // configure clocks
    RCC->APB2ENR |= (1 << 4);    // USART1EN = 1
    RCC->APB1ENR |= (1 << 17);   // USART2EN = 1
    RCC->APB1ENR |= (1 << 18);   // USART3EN = 1
    
    // configure pins PD8/PD9 as USART3 TX/RX
    setAlternateFunction(PORTD, 8, 7);
    setPinDir(PORTD, 9, GPIO_INPUT);
    
    // enable USART3
    USART3->CR1 |= (1 << 0);    // UE = 1
    
    // set baud rate to 57600 Baud at f_CPU = 216MHz, USART_clk = 27MHz
    USART3->BRR |= (0x1D5 << 0);    // USARTDIV = 0x1D5 = 469
    
    // configure USART3
    USART3->GTPR &= ~(0xFF << 0);   // PSC = 0
    USART3->GTPR |= (0x01 << 0);    // PSC = 1
    USART3->CR1 |= (1 << 5) | (1 << 3) | (1 << 2);    // RXNEIE = 1, TE = 1, RE = 1
    
    // Enable USART3 interrupt (NVIC level)
    NVIC->ISER[1] |= (1 << 7);    // enable USART3_IRQHandler interrupt (interrupt #39 which is bit 7 in ISER1 since 39 - 32 = 7)
}

void send_byte_UART(uint8_t byte)
{
    while(((USART3->ISR >> 7) & 0x01) == 0);    // wait until TXE = 1
    USART3->TDR = byte;
    while(((USART3->ISR >> 6) & 0x01) == 0);    // wait until TC = 1
}

void send_byte_hex_UART(uint8_t byte)
{
    send_byte_UART(hex_chars[(byte >> 4) & 0xF]);
    send_byte_UART(hex_chars[(byte >> 0) & 0xF]);
}

void send_string_number_UART(char *string, uint32_t number)
{
    send_string_UART(string);
    uint32_t nbr = number;    // maximum = 4294967295
    char buff[12];
    for(uint8_t i = 0; i < 10; i++)
    {
        buff[9 - i] = (nbr % 10) + 48;
        nbr = nbr / 10;
    }
    buff[10] = '\n';
    buff[11] = '\0';
    send_string_UART(buff);
}

void send_string_UART(char *string)
{
    char *ptr = string;
    
    while(*ptr != '\0')
    {
        send_byte_UART(*ptr);
        ptr++;
    }
}

void USART3_IRQHandler(void)
{
    USART3->RDR;    // we need to (at least) read the RDR register to clear the RXNE flag. This is done to avoid the USART to be stuck.
    
    // Do stuff ...
}
